package com.bcxin.tenant.open.dubbo.reader.providers.impls;

import com.bcxin.tenant.open.document.domains.documents.*;
import com.bcxin.tenant.open.document.domains.utils.KeywordTermUtils;
import com.bcxin.tenant.open.dubbo.reader.providers.transfers.DataTransfer;
import com.bcxin.tenant.open.infrastructure.redis.RedisConfig;
import com.bcxin.tenant.open.infrastructures.TenantContext;
import com.bcxin.tenant.open.infrastructures.TenantEmployeeContext;
import com.bcxin.tenant.open.infrastructures.constants.BusinessConstants;
import com.bcxin.tenant.open.infrastructures.enums.DispatchAccountType;
import com.bcxin.tenant.open.infrastructures.enums.RoleType;
import com.bcxin.tenant.open.infrastructures.exceptions.ArgumentTenantException;
import com.bcxin.tenant.open.infrastructures.exceptions.BadTenantException;
import com.bcxin.tenant.open.infrastructures.exceptions.NoFoundTenantException;
import com.bcxin.tenant.open.infrastructures.exceptions.UnAuthorizedTenantException;
import com.bcxin.tenant.open.infrastructures.utils.StringUtil;
import com.bcxin.tenant.open.jdks.DeviceReaderRpcProvider;
import com.bcxin.tenant.open.jdks.requests.DeviceSearchRequest;
import com.bcxin.tenant.open.jdks.responses.DeviceResponse;
import com.redis.om.spring.search.stream.EntityStream;
import com.redis.om.spring.search.stream.SearchStream;
import org.apache.dubbo.config.annotation.DubboService;
import org.springframework.data.geo.Distance;
import org.springframework.data.geo.Point;
import org.springframework.data.redis.domain.geo.Metrics;
import org.springframework.util.CollectionUtils;
import org.springframework.util.StringUtils;

import java.util.*;
import java.util.stream.Collectors;

@DubboService
public class DeviceReaderRpcProviderImpl implements DeviceReaderRpcProvider {
    private final EntityStream entityStream;
    private final DataTransfer dataTransfer;

    public DeviceReaderRpcProviderImpl(EntityStream entityStream, DataTransfer dataTransfer) {
        this.entityStream = entityStream;
        this.dataTransfer = dataTransfer;
    }

    @Override
    public Collection<DeviceResponse> search(DeviceSearchRequest request) {
        TenantEmployeeContext.TenantUserModel userModel =
                TenantContext.getInstance().getUserContext().get();
        if (userModel == null) {
            throw new UnAuthorizedTenantException();
        }

        /**
         * 根据监管归属进行过滤, 暂时不支持企业级的调度台过滤
         */
        SearchStream<RdDeviceDocument> deviceDocumentSearchStream =
                this.entityStream.of(RdDeviceDocument.class);
        RoleType roleType = RoleType.create(userModel.getOrgInstitutional(), userModel.getOrgName());
        if (userModel.isSuperviseDepartRole()) {
            String[] superviseDepartIds =
                    userModel.getMatchedSuperviseDepartIds(request.getSuperviseDepartIds())
                            .toArray(ii -> new String[ii]);
            if (superviseDepartIds.length == 0) {
                return Collections.EMPTY_LIST;
            }

            deviceDocumentSearchStream = deviceDocumentSearchStream
                    .filter(RdEmployeeDocument$.SCOPE_PERMISSIONS.in(superviseDepartIds));

        } else {
            /**
             * 支持本企业或者内保调度台调度
             */
            deviceDocumentSearchStream = deviceDocumentSearchStream
                    .filter(RdEmployeeDocument$.SCOPE_PERMISSIONS.in(userModel.getNoSuperviseScopePermissions()));
            if (roleType == RoleType.Company) {
                String[] excludeValues = new String[]{
                        BusinessConstants.DefaultEmptyValue
                };

                deviceDocumentSearchStream = deviceDocumentSearchStream
                        .filter(RdEmployeeDocument$.SECURITY_STATION_ID.notEq(excludeValues));
            }
        }

        deviceDocumentSearchStream = filterByLimitedStations(userModel, roleType, deviceDocumentSearchStream);

        boolean skipPage = request.isSkipPageLimit();
        switch (request.getSearchModel()) {
            case Geo -> {
                if(request.getLatitude()==null || request.getLongitude()==null || request.getDistance()==null) {
                    throw new ArgumentTenantException("周边搜索的时候, 经纬度及距离信息不能为空");
                }

                deviceDocumentSearchStream =
                        deviceDocumentSearchStream
                                .filter(RdEmployeeDocument$.LON_LAT.near(
                                        new Point(request.getLongitude(), request.getLatitude()),
                                        new Distance(request.getDistance(), Metrics.METERS)
                                ));

                skipPage = true;
            }
            case Keyword -> {
                if (!StringUtils.hasLength(request.getKeyword())) {
                    throw new BadTenantException("搜索关键字不可以为空");
                }

                deviceDocumentSearchStream =
                        filterByKeyword(deviceDocumentSearchStream, request.getKeyword());
            }
            case Selected -> {
                if (CollectionUtils.isEmpty(request.getIds())) {
                    return Collections.EMPTY_LIST;
                }

                String[] selectedIds = request.getIds().stream().filter(ii -> !StringUtil.isEmpty(ii))
                        .distinct().toArray(ix -> new String[ix]);
                if (selectedIds == null || selectedIds.length == 0) {
                    return Collections.EMPTY_LIST;
                }

                deviceDocumentSearchStream =
                        deviceDocumentSearchStream
                                .filter(RdDeviceDocument$.ID.in(selectedIds));
            }
            case SecurityResource -> {
                /**
                 * 辖区在上面已经进行过滤
                 */
                skipPage = true;
                if (!userModel.isSuperviseDepartRole() && !CollectionUtils.isEmpty(request.getSuperviseDepartIds())) {
                    deviceDocumentSearchStream = deviceDocumentSearchStream
                            .filter(RdEmployeeDocument$.SCOPE_PERMISSIONS.in(request.getSuperviseDepartIds().toArray(ii -> new String[ii])));
                }

                deviceDocumentSearchStream =
                        filterByKeyword(deviceDocumentSearchStream, request.getKeyword());
            }
        }

        int pageIndex = request.getPageIndex();
        if (pageIndex < 1) {
            pageIndex = 1;
        }

        Collection<RdDeviceDocument> documents = null;
        if (skipPage) {
            documents = deviceDocumentSearchStream
                    /**
                     * 这个一定要再map之前, 否则无法执行
                     */
                    .limit(RedisConfig.MAX_LIMIT)
                    .collect(Collectors.toList());
        } else {
            documents = deviceDocumentSearchStream.skip((pageIndex - 1) * request.getPageSize())
                    .limit(request.getPageSize()).collect(Collectors.toList());
        }

        return this.dataTransfer.translate(documents, request.getSearchModel() == DeviceSearchRequest.SearchModel.SecurityResource);
    }

    @Override
    public DeviceResponse get(String id) {
        RdDeviceDocument document =
                this.entityStream.of(RdDeviceDocument.class).filter(RdDeviceDocument$.ID.eq(id)).findFirst().orElse(null);

        if (document == null) {
            throw new NoFoundTenantException();
        }

        return this.dataTransfer.translate(document, false);
    }

    private static SearchStream<RdDeviceDocument> filterByLimitedStations(TenantEmployeeContext.TenantUserModel userModel,
                                                                            RoleType roleType,
                                                                            SearchStream<RdDeviceDocument> deviceDocumentSearchStream
    ) {

        /**
         * 内保单位的话; 则除了主管理员之外, 仅指定的保卫管理员可以查看到数据
         */
        if(roleType==RoleType.Proprietor) {
            if (userModel.getPermission() != null && userModel.getPermission().isDomainAdmin()) {
                return deviceDocumentSearchStream;
            }

            String[] limitedPermissionStationIds = userModel.getLimitedPermissionStations();
            deviceDocumentSearchStream = deviceDocumentSearchStream
                    .filter(RdDeviceDocument$.SECURITY_STATION_ID.in(limitedPermissionStationIds));

            return deviceDocumentSearchStream;
        }

        if (userModel.getAccountType() == DispatchAccountType.AppDesk) {

            /**
             * 企业/群防群治和派出所将限制
             * 社区警务的负责人仅查看到对应社区警务的人员信息
             * 驻勤点负责人仅查看对应的驻勤点的人员信息
             * */
            if (roleType == RoleType.Company ||
                    roleType == RoleType.PoliceStation ||
                    roleType == RoleType.PopulationPoliceStation) {
                String[] limitedPermissionStationIds = userModel.getLimitedPermissionStations();
                /**
                 * 针对企业及群防群治企业
                 */
                if (limitedPermissionStationIds != null &&
                        limitedPermissionStationIds.length > 0) {
                    if (roleType == RoleType.Company) {
                        deviceDocumentSearchStream = deviceDocumentSearchStream
                                .filter(RdDeviceDocument$.SECURITY_STATION_ID.in(userModel.getLimitedPermissionStations()));
                    } else {
                        String filteredValues = Arrays.stream(limitedPermissionStationIds)
                                .collect(Collectors.joining(" | "));
                        deviceDocumentSearchStream = deviceDocumentSearchStream
                                .filter(String.format("(@resourceTypes:{Population} @securityStationId:{%s}) | -(@resourceTypes:{Population})", filteredValues));
                    }
                }
            }

        }

        return deviceDocumentSearchStream;
    }

    private static SearchStream<RdDeviceDocument> filterByKeyword(SearchStream<RdDeviceDocument> deviceDocumentSearchStream,
                                                                    String keyword) {
        if (StringUtils.hasLength(keyword)) {
            String searchKeyword = keyword.trim();
            String expression = KeywordTermUtils.getFilterExpression(searchKeyword);
            deviceDocumentSearchStream = deviceDocumentSearchStream.filter(expression);
        }

        return deviceDocumentSearchStream;
    }
}
