package com.bcxin.tenant.open.dubbo.reader.providers.impls;

import com.bcxin.tenant.open.document.domains.documents.RdEmployeeDocument$;
import com.bcxin.tenant.open.document.domains.documents.RdSecurityStationDocument;
import com.bcxin.tenant.open.document.domains.documents.RdSecurityStationDocument$;
import com.bcxin.tenant.open.document.domains.repositories.RdSecurityStationDocumentRepository;
import com.bcxin.tenant.open.document.domains.utils.KeywordTermUtils;
import com.bcxin.tenant.open.dubbo.reader.providers.transfers.DataTransfer;
import com.bcxin.tenant.open.infrastructure.redis.RedisConfig;
import com.bcxin.tenant.open.infrastructures.EntityCollection;
import com.bcxin.tenant.open.infrastructures.TenantContext;
import com.bcxin.tenant.open.infrastructures.TenantEmployeeContext;
import com.bcxin.tenant.open.infrastructures.constants.BusinessConstants;
import com.bcxin.tenant.open.infrastructures.enums.DeskType;
import com.bcxin.tenant.open.infrastructures.enums.DispatchAccountType;
import com.bcxin.tenant.open.infrastructures.enums.ResourceType;
import com.bcxin.tenant.open.infrastructures.enums.RoleType;
import com.bcxin.tenant.open.infrastructures.exceptions.ArgumentTenantException;
import com.bcxin.tenant.open.infrastructures.exceptions.NoAllowedTenantException;
import com.bcxin.tenant.open.infrastructures.exceptions.UnAuthorizedTenantException;
import com.bcxin.tenant.open.infrastructures.utils.BusinessUtil;
import com.bcxin.tenant.open.jdks.SecurityStationReaderRpcProvider;
import com.bcxin.tenant.open.jdks.requests.SecurityStationReaderRequest;
import com.bcxin.tenant.open.jdks.requests.SecurityStationShortSearchRequest;
import com.bcxin.tenant.open.jdks.requests.StationStatisticsSearchRequest;
import com.bcxin.tenant.open.jdks.responses.SecurityStationDetailResponse;
import com.bcxin.tenant.open.jdks.responses.SecurityStationOverviewReaderResponse;
import com.bcxin.tenant.open.jdks.responses.SecurityStationShortResponse;
import com.bcxin.tenant.open.jdks.responses.StationStatisticsResponse;
import com.redis.om.spring.search.stream.EntityStream;
import com.redis.om.spring.search.stream.SearchStream;
import org.apache.dubbo.config.annotation.DubboService;
import org.springframework.data.geo.Distance;
import org.springframework.data.geo.Point;
import org.springframework.data.redis.domain.geo.Metrics;
import org.springframework.util.CollectionUtils;
import org.springframework.util.StringUtils;

import java.util.*;
import java.util.stream.Collectors;

@DubboService
public class SecurityStationReaderRpcProviderImpl implements SecurityStationReaderRpcProvider {
    private final EntityStream entityStream;
    private final DataTransfer dataTransfer;
    private final RdSecurityStationDocumentRepository rdSecurityStationDocumentRepository;

    public SecurityStationReaderRpcProviderImpl(EntityStream entityStream, DataTransfer dataTransfer,
                                                RdSecurityStationDocumentRepository rdSecurityStationDocumentRepository) {
        this.entityStream = entityStream;
        this.dataTransfer = dataTransfer;
        this.rdSecurityStationDocumentRepository = rdSecurityStationDocumentRepository;
    }

    @Override
    public Collection<SecurityStationOverviewReaderResponse> search(SecurityStationReaderRequest request) {

        if (CollectionUtils.isEmpty(request.getDeskTypes())) {
            throw new ArgumentTenantException("参数异常; 工作台类型(deskTypes)不能为空");
        }

        TenantEmployeeContext.TenantUserModel userModel = TenantContext.getInstance().getUserContext().get();
        if (userModel == null) {
            throw new UnAuthorizedTenantException();
        }

        /**
         * 临保调度按照安保资源进行展示
         */
        SecurityStationReaderRequest.SearchModel searchModel = request.getSearchModel();
        boolean isTemporarySecurity = false;
        if (DeskType.hasDeskType(request.getDeskTypes(), DeskType.TemporarySecurity) &&
                userModel.getAccountType()==DispatchAccountType.Desk &&
                searchModel== SecurityStationReaderRequest.SearchModel.Name) {
            searchModel = SecurityStationReaderRequest.SearchModel.SecurityResource;
            isTemporarySecurity = true;
        }

        SearchStream<RdSecurityStationDocument> searchStream =
                this.entityStream.of(RdSecurityStationDocument.class);

        Set<String> selectedResources
                = ResourceType.getResources(request.getDeskTypes());
        if (!CollectionUtils.isEmpty(selectedResources)) {
            searchStream =
                    searchStream.filter(RdSecurityStationDocument$.RESOURCE_TYPES.in(
                            selectedResources.toArray(ix -> new String[ix])
                    ));
        }

        if (request.getResourceType() != null) {
            searchStream = searchStream.filter(RdSecurityStationDocument$.RESOURCE_TYPES.in(request.getResourceType().name()));
        }

        if (searchModel != SecurityStationReaderRequest.SearchModel.Organization) {
            if (userModel.isSuperviseDepartRole()) {
                String[] superviseDepartIds =
                        userModel.getMatchedSuperviseDepartIds(request.getSuperviseDepartIds())
                                .toArray(ii -> new String[ii]);

                if (superviseDepartIds.length == 0) {
                    return Collections.EMPTY_LIST;
                }

                /**
                 * 根据监管归属进行过滤, 暂时不支持企业级的调度台过滤
                 */
                searchStream = searchStream.filter(
                        RdSecurityStationDocument$.SCOPE_PERMISSIONS
                                .in(superviseDepartIds)
                );
            } else {
                /**
                 * 支持本企业（及子公司调度）调度台调度
                 */

                Collection<String> departIds = new HashSet<>();
                departIds.addAll(Arrays.stream(userModel.getNoSuperviseScopePermissions()).toList());
                if (!userModel.isSuperviseDepartRole() && !CollectionUtils.isEmpty(request.getSuperviseDepartIds())) {
                    departIds.addAll(request.getSuperviseDepartIds());
                }

                searchStream = searchStream
                        .filter(RdSecurityStationDocument$.SCOPE_PERMISSIONS.in(departIds.toArray(ix -> new String[ix])));
            }

            RoleType roleType = RoleType.create(userModel.getOrgInstitutional(), userModel.getOrgName());
            searchStream = filterByLimitedStations(userModel, roleType, searchStream);
        }

        if (StringUtils.hasLength(request.getProjectId())) {
            searchStream = searchStream.filter(RdSecurityStationDocument$.SCOPE_PERMISSIONS.in(request.getProjectId()));
        }

        if (!CollectionUtils.isEmpty(request.getStationTypes())) {
            String[] stationTypes = request.getStationTypes().toArray(new String[request.getStationTypes().size()]);

            searchStream = searchStream.filter(RdSecurityStationDocument$.STATION_TYPES.in(stationTypes));
        }

        /**
         * https://github.com/redis-developer/redis-om-spring-skeleton-app
         */
        boolean skipPage = false;
        switch (searchModel) {
            case Location -> {
                skipPage = true;
                searchStream = searchStream.filter(RdSecurityStationDocument$.LON_LAT.near(
                        new Point(request.getLongitude(), request.getLatitude()),
                        new Distance(request.getDistance(), Metrics.METERS)
                ));
            }
            case SecurityResource -> {
                skipPage = true;
                if (request.getFollowed() != null) {
                    if (request.getFollowed()) {
                        searchStream =
                                searchStream.filter(
                                        RdEmployeeDocument$.FOLLOWED_DEVICE_NOS.in(userModel.getTencentUserId())
                                );
                    } else {
                        /**
                         * 关注人不在列表中
                         */
                        Set<String> noDeviceNos = new HashSet<>();
                        noDeviceNos.add(userModel.getTencentUserId());

                        searchStream =
                                searchStream.filter(
                                        RdEmployeeDocument$.FOLLOWED_DEVICE_NOS.containsNone(noDeviceNos)
                                );
                    }
                }

                searchStream = filterByKeyword(searchStream, request.getKeyword());
            }
            /**
             * 调度通讯和指挥调度均支持通过姓名进行搜索
             */
            case Name, DispatchBookAddress -> {
                searchStream = filterByKeyword(searchStream, request.getKeyword());
            }
            case Organization -> {
                if (!StringUtils.hasLength(request.getOrganizationId())) {
                    throw new NoAllowedTenantException("企业Id不允许为空");
                }

                //searchStream = searchStream.filter(RdSecurityStationDocument$.ORGANIZATION_ID.eq(request.getOrganizationId()));
                /**
                 * 这边使用这个的原因是内保单位查询驻勤点详情的时候; 需要根据有权限的企业可以查得到该驻勤列表
                 */
                searchStream = searchStream.filter(RdSecurityStationDocument$.SCOPE_PERMISSIONS.in(request.getOrganizationId()));
            }
            /**
             * 比如： 驻勤点类型
             */
            default -> {
                if (!CollectionUtils.isEmpty(request.getStationTypes())) {
                    String[] stationTypes = request.getStationTypes().toArray(new String[request.getStationTypes().size()]);


                    searchStream = searchStream.filter(RdSecurityStationDocument$.STATION_TYPES.in(stationTypes));
                    skipPage = true;
                }
            }
        }

        int pageIndex = request.getPageIndex();
        if (pageIndex < 1) {
            pageIndex = 1;
        }

        Collection<RdSecurityStationDocument> documents = null;
        if (skipPage) {
            documents = searchStream
                    /**
                     * 这个一定要再map之前, 否则无法执行
                     */
                    .limit(RedisConfig.MAX_LIMIT)
                    .collect(Collectors.toList());
        } else {
            documents = searchStream.skip((pageIndex - 1) * request.getPageSize())
                    .limit(request.getPageSize()).collect(Collectors.toList());
        }

        if (isTemporarySecurity) {
            searchModel = SecurityStationReaderRequest.SearchModel.Name;
        }

        return this.dataTransfer.translate2Stations(
                searchModel,
                documents);
    }

    @Override
    public SecurityStationDetailResponse get(String id) {
        Optional<RdSecurityStationDocument> documentOptional =
                this.entityStream
                        .of(RdSecurityStationDocument.class)
                        .filter(RdSecurityStationDocument$.ID.eq(id))
                        .findFirst();

        if (!documentOptional.isPresent()) {
            return null;
        }

        return this.dataTransfer.translate(this.entityStream, documentOptional.get());
    }

    @Override
    public Collection<String> getSecurityStationIdsByOrganizationIds(Collection<String> organizationIds) {
        if (CollectionUtils.isEmpty(organizationIds)) {
            return Collections.emptyList();
        }

        String[] organizationIdsArray = organizationIds.toArray(new String[organizationIds.size()]);

        Collection<String> ids =
                this.entityStream.of(RdSecurityStationDocument.class)
                        .filter(RdSecurityStationDocument$.ORGANIZATION_ID.in(organizationIdsArray))
                        .collect(Collectors.toList()).stream()
                        .map(ix -> ix.getId()).collect(Collectors.toList());

        return ids;
    }

    @Override
    public Collection<String> getOrganizationIdsByStationIds(Collection<String> stationIds) {
        if (CollectionUtils.isEmpty(stationIds)) {
            return Collections.emptyList();
        }

        String[] stationIdsArray = stationIds.toArray(new String[stationIds.size()]);

        Collection<String> ids =
                this.entityStream.of(RdSecurityStationDocument.class)
                        .filter(RdSecurityStationDocument$.ID.in(stationIdsArray))
                        .collect(Collectors.toList()).stream()
                        .map(ix -> ix.getOrganizationId())
                        .distinct().collect(Collectors.toList());

        return ids;
    }

    @Override
    public Collection<StationStatisticsResponse> search(StationStatisticsSearchRequest request) {
        TenantEmployeeContext.TenantUserModel userModel = TenantContext.getInstance().getUserContext().get();
        if (userModel == null) {
            throw new UnAuthorizedTenantException();
        }

        SearchStream<RdSecurityStationDocument> searchStream =
                this.entityStream.of(RdSecurityStationDocument.class);
        if (userModel.isSuperviseDepartRole()) {
            searchStream = searchStream.filter(RdSecurityStationDocument$.SCOPE_PERMISSIONS.in(userModel.getAssignedSuperviseDepartIds()));
        } else {
            /**
             * 本公司及子公司的数据查询
             */
            searchStream = searchStream.filter(RdSecurityStationDocument$.SCOPE_PERMISSIONS.in(userModel.getNoSuperviseScopePermissions()));
        }

        if (!CollectionUtils.isEmpty(request.getStationIds())) {
            String[] selectedStationIds = request.getStationIds().toArray(ix -> new String[ix]);
            searchStream = searchStream.filter(RdSecurityStationDocument$.ID.in(selectedStationIds));
        }

        if (!CollectionUtils.isEmpty(request.getStationTypeCode())) {
            String[] selectedStationTypes = request.getStationTypeCode().toArray(ix -> new String[ix]);
            searchStream = searchStream.filter(RdSecurityStationDocument$.STATION_TYPES.in(selectedStationTypes));
        }

        Collection<RdSecurityStationDocument> documents = searchStream
                /**
                 * 这个一定要再map之前, 否则无法执行
                 */
                .limit(RedisConfig.MAX_LIMIT)
                .collect(Collectors.toList());

        Collection<String> stationTypes =
                documents.stream()
                        .filter(ii -> !CollectionUtils.isEmpty(ii.getStationTypes()))
                        .flatMap(ii -> ii.getStationTypes().stream())
                        .distinct().collect(Collectors.toSet());

        Collection<StationStatisticsResponse> statisticsResponses =
                stationTypes.stream().map(ii -> {
                    Collection<RdSecurityStationDocument> selectedStations =
                            documents.stream().filter(ix -> ix.getStationTypes() != null && ix.getStationTypes().contains(ii))
                                    .collect(Collectors.toList());

                    int countOfSecurityMen = selectedStations.stream().collect(Collectors.summingInt(ix -> ix.getCountOfSecurityMan()));
                    int countOfStation = (int) selectedStations.stream().count();
                    return StationStatisticsResponse.create(ii, countOfStation, countOfSecurityMen);
                }).collect(Collectors.toList());

        return statisticsResponses;
    }

    @Override
    public EntityCollection<SecurityStationShortResponse> search(SecurityStationShortSearchRequest request) {

        TenantEmployeeContext.TenantUserModel userModel = TenantContext.getInstance().getUserContext().get();
        if (userModel == null) {
            throw new UnAuthorizedTenantException();
        }

        SearchStream<RdSecurityStationDocument> searchStream = getSearchStream(
                userModel, request.getSuperviseDepartIds()
        );

        if (StringUtils.hasLength(request.getName())) {
            searchStream = filterByKeyword(searchStream, request.getName());
        }

        if (StringUtils.hasLength(request.getCompanyName())) {
            searchStream = searchStream.filter(RdSecurityStationDocument$.COMPANY_NAME.startsWith(request.getCompanyName().trim()));
        }

        if(!CollectionUtils.isEmpty(request.getStationTypes())) {
            String[] stationTypes = request.getStationTypes().toArray(ix -> new String[ix]);
            searchStream = searchStream.filter(RdSecurityStationDocument$.STATION_TYPES.in(stationTypes));
        }

        int pageIndex = request.getPageIndex();
        if (pageIndex < 1) {
            pageIndex = 1;
        }

        long totalCount = searchStream.count();
        Collection<RdSecurityStationDocument>
                documents = searchStream.skip((pageIndex - 1) * request.getPageSize())
                .limit(request.getPageSize())
                .collect(Collectors.toList());

        Collection<SecurityStationShortResponse> shortResponses =
                documents.stream().map(ii ->
                                SecurityStationShortResponse.create(ii.getId(), ii.getName(), ii.getCompanyName(), ii.getScopePermissions(),ii.getStationTypes()))
                        .collect(Collectors.toList());

        return EntityCollection.create(shortResponses, request.getPageSize(), totalCount);
    }

    private static SearchStream<RdSecurityStationDocument> filterByKeyword(
            SearchStream<RdSecurityStationDocument> searchStream,
            String input) {
        if (StringUtils.hasLength(input)) {
            String searchKeyword = input.trim();
            String expression = KeywordTermUtils.getFilterExpression(searchKeyword);
            searchStream = searchStream.filter(expression);
        }

        return searchStream;
    }

    private static SearchStream<RdSecurityStationDocument> filterByLimitedStations(TenantEmployeeContext.TenantUserModel userModel,
                                                                            RoleType roleType,
                                                                            SearchStream<RdSecurityStationDocument> securityStationDocumentSearchStream
    ) {

        /**
         * 内保单位的话; 则除了主管理员之外, 仅指定的保卫管理员可以查看到数据
         */
        if(roleType==RoleType.Proprietor) {
            if (userModel.getPermission() != null && userModel.getPermission().isDomainAdmin()) {
                return securityStationDocumentSearchStream;
            }

            String[] limitedPermissionStationIds = userModel.getLimitedPermissionStations();
            securityStationDocumentSearchStream = securityStationDocumentSearchStream
                    .filter(RdSecurityStationDocument$.ID.in(limitedPermissionStationIds));

            return securityStationDocumentSearchStream;
        }

        /**
         * 仅移动端的调度台做如下的限制
         */
        if (userModel.getAccountType() == DispatchAccountType.AppDesk ||
                /**
                 * 针对赛演的管理单位, 我们限制其单位人员可以访问的权限
                 */
                BusinessUtil.isDaHuoManagedOrg(userModel.getOrgInstitutional())) {
            /**
             * 企业/群防群治和派出所将限制
             * 社区警务的负责人仅查看到对应社区警务
             * 驻勤点负责人仅查看对应的驻勤点
             * */
            if (roleType == RoleType.Company ||
                    roleType == RoleType.PoliceStation ||
                    roleType == RoleType.PopulationPoliceStation) {
                Set<String> limitedValidPermissionStationIds =
                        Arrays.stream(userModel.getLimitedPermissionStations())
                                /**
                                 * 过滤非项目信息
                                 */
                        .filter(ii-> !BusinessConstants.DefaultEmptyValue.equalsIgnoreCase(ii) && !BusinessConstants.isRelativeProjectIdStuffId(ii))
                        .collect(Collectors.toSet());
                /**
                 * 针对企业及群防群治企业
                 */
                if (!CollectionUtils.isEmpty(limitedValidPermissionStationIds)) {
                    if (roleType == RoleType.Company) {
                        if(!userModel.isDomainAdmin()) {
                            String[] limitedPermissionStationIds = limitedValidPermissionStationIds.toArray(ix -> new String[ix]);
                            securityStationDocumentSearchStream = securityStationDocumentSearchStream
                                    .filter(RdSecurityStationDocument$.ID.in(limitedPermissionStationIds));
                        }else {
                            /**
                             * 限制指定的岗点或者可查看本企业的所有数据
                             */
                            String filteredValues = limitedValidPermissionStationIds.stream().collect(Collectors.joining(" | "));
                            securityStationDocumentSearchStream = securityStationDocumentSearchStream
                                    .filter(String.format("(@id:{%s}) | (@scopePermissions:{%s})", filteredValues, userModel.getOrganizationId()));
                        }
                    } else {
                        String filteredValues = limitedValidPermissionStationIds.stream().collect(Collectors.joining(" | "));
                        securityStationDocumentSearchStream = securityStationDocumentSearchStream
                                .filter(String.format("(@resourceTypes:{Population} @id:{%s}) | -(@resourceTypes:{Population})", filteredValues));
                    }
                }
            }
        }

        return securityStationDocumentSearchStream;
    }

    private SearchStream<RdSecurityStationDocument> getSearchStream(TenantEmployeeContext.TenantUserModel userModel,
                                                                           Collection<String> selectedSuperviseDepartIds) {
        SearchStream<RdSecurityStationDocument> searchStream =
                this.entityStream.of(RdSecurityStationDocument.class);

        if (userModel.isSuperviseDepartRole()) {
            String[] superviseDepartIds =
                    userModel.getMatchedSuperviseDepartIds(selectedSuperviseDepartIds)
                            .toArray(ii -> new String[ii]);

            /**
             * 根据监管归属进行过滤, 暂时不支持企业级的调度台过滤
             */
            searchStream = searchStream.filter(
                    RdSecurityStationDocument$.SCOPE_PERMISSIONS
                            .in(superviseDepartIds)
            );
        } else {
            /**
             * 支持本企业调度台调度
             */

            searchStream = searchStream
                    .filter(RdEmployeeDocument$.SCOPE_PERMISSIONS.in(userModel.getOrganizationId()));

            if (!userModel.isSuperviseDepartRole() && !CollectionUtils.isEmpty(selectedSuperviseDepartIds)) {
                searchStream = searchStream
                        .filter(RdEmployeeDocument$.SCOPE_PERMISSIONS.in(selectedSuperviseDepartIds.toArray(ix -> new String[ix])));
            }


        }

        return searchStream;
    }
}
